package org.maachang.dao ;

import org.maachang.dao.dbms.Record;
import org.maachang.dao.dbms.RecordFactory;

/**
 * DaoSessionFactory.
 *
 * @version 2007/11/03
 * @author  masahito suzuki
 * @since MaachangDao 1.00
 */
public class DaoSessionFactory {
    
    /**
     * シングルトン.
     */
    private static final DaoSessionFactory SNGL = new DaoSessionFactory() ;
    
    /**
     * レコードファクトリ.
     */
    private RecordFactory recordFactory = null ;
    
    /**
     * セッション管理情報.
     */
    private DaoSession[] daoSessions = null ;
    
    /**
     * コンストラクタ.
     */
    private DaoSessionFactory() {
    }
    
    /**
     * オブジェクトを取得.
     * <BR><BR>
     * オブジェクトを取得します.
     * <BR>
     * @return DaoSessionFactory オブジェクト情報が返されます.
     */
    public static final DaoSessionFactory getInstance() {
        return SNGL ;
    }
    
    /**
     * レコードファクトリを設定.
     * <BR><BR>
     * レコードファクトリを設定します.
     * <BR>
     * @param recordFactory 対象のレコードファクトリを設定します.
     */
    protected synchronized void setRecordFactory( RecordFactory recordFactory ) {
        this.recordFactory = recordFactory ;
        int len = recordFactory.getConnectionManager().getMax() ;
        this.daoSessions = new DaoSession[ len ] ;
        for( int i = 0 ; i < len ; i ++ ) {
            this.daoSessions[ i ] = new DaoSession() ;
        }
    }
    
    /**
     * レコードファクトリを取得.
     * @return RecordFactory レコードファクトリが返されます.
     */
    public synchronized RecordFactory getRecordFactory() {
        return recordFactory ;
    }
    
    /**
     * 新しいDaoSessionを取得.
     * <BR><BR>
     * 新しいDaoSessionを取得します.
     * <BR>
     * @return DaoSession 新しいDaoSessionが返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public DaoSession getDaoSession()
        throws MaachangDaoException {
        if( recordFactory != null ) {
            Record record = null ;
            DaoSession ret = null ;
            try {
                record = recordFactory.getRecord() ;
                ret = poolDaoSession( record ) ;
            } catch( Exception e ) {
                if( record != null ) {
                    try {
                        record.close() ;
                    } catch( Exception ee ) {
                    }
                }
                throw new MaachangDaoException( e ) ;
            }
            return ret ;
        }
        throw new MaachangDaoException(
            "初期化されていないか、条件が不正です" ) ;
    }
    
    /**
     * 新しいDaoSessionを取得.
     * <BR><BR>
     * 新しいDaoSessionを取得します.
     * <BR>
     * @param record 対象のレコードを設定します.
     * @return DaoSession 新しいDaoSessionが返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public DaoSession getDaoSession( Record record )
        throws MaachangDaoException {
        DaoSession s = poolDaoSession( record ) ;
        return s ;
    }
    
    /**
     * 新しいレコードを取得.
     * <BR><BR>
     * 新しいレコードを取得します.
     * <BR>
     * @return Record 新しいレコード情報が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public Record getRecord()
        throws MaachangDaoException {
        if( recordFactory != null ) {
            Record ret = null ;
            try {
                ret = recordFactory.getRecord() ;
            } catch( Exception e ) {
                if( ret != null ) {
                    try {
                        ret.close() ;
                    } catch( Exception ee ) {
                    }
                }
                throw new MaachangDaoException( e ) ;
            }
            return ret ;
        }
        throw new MaachangDaoException(
            "初期化されていないか、条件が不正です" ) ;
    }
    
    /**
     * １つのDaoSessionオブジェクトを取得.
     */
    private synchronized DaoSession poolDaoSession( Record record ) {
        int len = daoSessions.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            DaoSession s = daoSessions[ i ] ;
            if( s.isUsedObject() == false ) {
                s.create( record ) ;
                return s ;
            }
        }
        return new DaoSession() ;
    }
}
